/*
 * (c) 2022, SWD Embedded Systems Limited, http://www.kpda.ru
 */


#include "externs.h"
#include <sys/mman.h>


void set_port( unsigned port, unsigned mask, unsigned data )
{
    unsigned char   c;

    c = read_8250( port );
    write_8250( port, (c & ~mask) | (data & mask) );
}


static void clear_device( const uintptr_t *port )
{
    unsigned char   tmp __attribute__(( unused ));

    write_8250( port[REG_IE], IE_CLR_ALL ); /* Disable all interrupts */
    tmp = read_8250( port[REG_LS] );        /* Clear Line Status Interrupt */
    tmp = read_8250( port[REG_RX] );        /* Clear RX Interrupt */
    tmp = read_8250( port[REG_TX] );        /* Clear TX Interrupt */
    tmp = read_8250( port[REG_MS] );        /* Clear Modem Interrupt */
    write_8250( port[REG_FC], 0x00 );       /* Disable and reset FIFOs */
}


/* Clean up the device then add it to the interrupt list and enable it. */
int ser_attach_intr( DEV_8250 *dev )
{
    uintptr_t       *port = dev->port;
    struct dev_list **owner;
    struct dev_list *curr;

    /* According to the National bug sheet you must wait for the transmit holding register to be empty. */
    devc_dbg( "Wait TX empty" );
    do {
    } while ((read_8250( port[REG_LS] ) & LSR_TXRDY) == 0);

    /* Clean the device so we get a level change on the intr line to the bus. Enable out2 (gate intr to bus) */
    set_port( port[REG_MC], MCR_OUT2, MCR_OUT2 );
    clear_device( port );

    /* Add it to the interrupt list */
    owner = &devices;
    for ( ;; )
    {
        curr = *owner;
        if ( curr == NULL )
        {
            curr = malloc( sizeof( *curr ) );
            if ( curr == NULL )
            {
                devc_err( "io-char: Allocation of device interrupt entry failed (%d)", errno );
                return (-1);
            }
            *owner       = curr;
            curr->next   = NULL;
            curr->device = NULL;
            break;
        }
        if ( curr->device->intr == dev->intr )
            break;
        owner = &curr->next;
    }

    /* Delay interrupts while we're fiddling around with the list */
    InterruptMask( dev->intr, -1 );
    dev->next    = curr->device;
    curr->device = dev;
    InterruptUnmask( dev->intr, -1 );

    /* If first handler, attach the interrupt */
    if ( curr->device->next == NULL )
        curr->iid = InterruptAttach( dev->intr, ser_intr, curr, 0, 0 );

    /* Enable ALL interrupt sources */
    if ( dev->tty.c_cflag & CREAD )
        write_8250( port[REG_IE], IE_SET_ALL );
    else
        write_8250( port[REG_IE], (IE_SET_ALL & ~IE_RX) );
    devc_dbg( "Interrupt handler attached successfully" );

    return (EOK);
}


void ser_detach_intr( DEV_8250 *dev )
{
    struct dev_list *curr_list,
                    *prev_list;
    DEV_8250        *curr_dev,
                    *prev_dev;
    uintptr_t       *port = dev->port;

    /* Disable ALL interrupt sources */
    write_8250( port[REG_IE], IE_CLR_ALL );     /* Disable interrupts */
    set_port( port[REG_MC], MCR_OUT2, 0x00 );   /* Disable out2 */

    /* Remove from list of devices to scan on an interrupt */

    /* Delay interrupts while we're fiddling around with the list */
    InterruptMask( dev->intr, -1 );

    /* Find the right interrupt list */
    curr_list = prev_list = devices;
    while ( curr_list )
    {
        if ( curr_list->device->intr == dev->intr )
        {
            /* Now that we found the right interrupt list find and remove the device from the
             * device list (but don't free it) */
            prev_dev = curr_dev = curr_list->device;
            while ( curr_dev )
            {
                if ( curr_dev == dev )
                {
                    if ( curr_dev == curr_list->device )    /* If first entry in device list */
                    {
                        curr_list->device = curr_list->device->next;
                        if ( curr_list->device == NULL )    /* No more devices in list */
                        {
                            if ( curr_list == devices )     /* If first list in list of lists :-) */
                                devices = devices->next;
                            else
                                prev_list->next = curr_list->next;
                            InterruptDetach( curr_list->iid );
                            free( curr_list );              /* Free empty list */
                            curr_list = NULL;
                        }
                    } else
                        prev_dev->next = curr_dev->next;

                    curr_dev = NULL;
                    curr_list = NULL;   /* We found what we were looking for, now bail out of both loops */
                } else {
                    if ( prev_dev != curr_dev )
                        prev_dev = prev_dev->next;
                    curr_dev = curr_dev->next;
                }
            }
        } else {
            if ( prev_list != curr_list )
                prev_list = prev_list->next;
            curr_list = curr_list->next;
        }
    }
    clear_device( dev->port );
    InterruptUnmask( dev->intr, -1 );

    dev->intr = _NTO_INTR_SPARE;

    return;
}


DEV_8250 * create_device( TTYINIT *dip, unsigned unit, exar_dev_t *exdev )
{
    DEV_8250        *dev;
    unsigned        i;
    uintptr_t       port;

    /* Get a device entry and the input/output buffers for it */
    if ( (dev = (void *)malloc( sizeof( *dev ) )) == NULL )
    {
        devc_err( "io-char: Allocation of device entry failed (%d)", errno );
        return (dev);
    }
    memset( dev, 0, sizeof( *dev ) );

    /* Get buffers */
    dev->tty.ibuf.size = dip->isize;
    dev->tty.ibuf.head = dev->tty.ibuf.tail = dev->tty.ibuf.buff = malloc( dev->tty.ibuf.size );
    if ( dev->tty.ibuf.head == NULL )
    {
        devc_err( "io-char: Allocation of input buffer failed (%d)", errno );
        free( dev );
        return (NULL);
    }

    dev->tty.obuf.size = dip->osize;
    dev->tty.obuf.head = dev->tty.obuf.tail = dev->tty.obuf.buff = malloc( dev->tty.obuf.size );
    if ( dev->tty.obuf.head == NULL )
    {
        devc_err( "io-char: Allocation of output buffer failed (%d)", errno );
        free( dev->tty.ibuf.head );
        free( dev );
        return (NULL);
    }

    dev->tty.cbuf.size = dip->csize;
    dev->tty.cbuf.head = dev->tty.cbuf.tail = dev->tty.cbuf.buff = malloc( dev->tty.cbuf.size );
    if ( dev->tty.cbuf.head == NULL )
    {
        devc_err( "io-char: Allocation of canonical buffer failed (%d)", errno );
        free( dev->tty.ibuf.head );
        free( dev->tty.obuf.head );
        free( dev );
        return (NULL);
    }

    if ( dip->highwater )
        dev->tty.highwater = dip->highwater;
    else
        dev->tty.highwater = dev->tty.ibuf.size - FIFO_SIZE * 2;

    if ( dev->tty.highwater <= 0 )
        return (NULL);

    strcpy( dev->tty.name, dip->name );

    dev->chnum = unit;
    if ( exar_dev.ch_offset < 0 )
        exar_dev.ch_offset = unit - 1;
    dev->tty.baud = dip->baud;
    dev->tty.fifo = dip->fifo;

#if defined( __X86__ )
    port = (uintptr_t)mmap_device_memory( NULL, (sizeof( dev->port ) / sizeof( dev->port[0] )) << dip->port_shift,
                                          PROT_READ | PROT_WRITE | PROT_NOCACHE, 0, dip->port );
#else
    port = mmap_device_io( (sizeof( dev->port ) / sizeof( dev->port[0] )) << dip->port_shift, dip->port );
#endif

    for ( i = 0; i < sizeof( dev->port ) / sizeof( dev->port[0] ); ++i )
    {
        dev->port[i]  = port;
        port         += 1 << dip->port_shift;
    }

    write_8250( dev->port[REG_EX_RESET], 0xFF );

    dev->intr        = dip->intr;
    dev->clk         = dip->clk;
    dev->div         = dip->div;
    dev->tty.flags   = EDIT_INSERT | LOSES_TX_INTR;
    dev->tty.c_cflag = dip->c_cflag;
    dev->tty.c_iflag = dip->c_iflag;
    dev->tty.c_lflag = dip->c_lflag;
    dev->tty.c_oflag = dip->c_oflag;
    dev->tty.verbose = dip->verbose;

    /* Initialize termios cc codes to an ANSI terminal */
    ttc( TTC_INIT_CC, &dev->tty, 0 );

    /* Initialize the device's name. Assume that the basename is set in device name. This will
     * attach to the path assigned by the unit number/minor number combination */
    unit = SET_NAME_NUMBER( unit ) | NUMBER_DEV_FROM_USER;
    ttc( TTC_INIT_TTYNAME, &dev->tty, unit );

    /* Initialize power management structures before attaching ISR */
    ttc( TTC_INIT_POWER, &dev->tty, 0 );

    /* Disable Interrupts on port. Interrupts must be disbled on chip for all ports on a multiport card
     * before attaching an ISR to any of the ports. enable_device() must be called for each port before
     * returning from options() to complete port initialization */
    clear_device( dev->port );

    dev->parent_dev = exdev;

    return (dev);
}


void free_device( DEV_8250 *dev )
{
    free( dev->tty.ibuf.head );
    free( dev->tty.obuf.head );
    free( dev->tty.cbuf.head );
    ser_detach_intr( dev );
    free( dev );
}


void enable_exar_ch( DEV_8250 *dev )
{
    uint8_t     fctr = read_8250( dev->port[REG_FCTR] );

    devc_dbg( "[CH%d] RTS/DTS : 0x%lX , TRG Table: 0x%X , SPR : 0x%X", dev->chnum,
              (fctr & 0xF), ((fctr >> 6) & 0x3), read_8250( dev->port[REG_SPR] ) );

    devc_dbg( "[CH%d] Invert IR input %s , Auto RS485 %s", dev->chnum,
              ((FIELD_VAL( fctr, REG_FCTR_INVERT_IR )) ? "enable" : "disable"),
              ((FIELD_VAL( fctr , REG_FCTR_AUTO_RS485 )) ? "enable" : "disable") );

    uint8_t efr = read_8250( dev->port[REG_EFR] );
    devc_dbg( "[CH%d] Software Flow cntl : 0x%X, Spec char select %s ", dev->chnum,
              (efr & 0xf), (((efr >> 5) & 1) ? "enable" : "disable") );
    devc_dbg( "[CH%d] Auto RTS/DTR %s , Auto CTS/DSR %s", dev->chnum,
              (((efr >> 6) & 1) ? "enable" : "disable"),
              (((efr >> 7) & 1) ? "enable" : "disable") );

    devc_dbg( "[CH%d] RX cntr : %d , TX cntr : %d", dev->chnum, read_8250( dev->port[REG_RXCNT] ),
              read_8250( dev->port[REG_TXCNT] ) );

    /* Restore default state */
    write_8250( dev->port[REG_SPR], 0xFF );
    write_8250( dev->port[REG_FCTR], 0 );
    write_8250( dev->port[REG_EFR], 0 );

    devc_dbg( "[CH%d] INT regs : 0x%X 0x%X 0x%X 0x%X", dev->chnum,  read_8250( dev->port[REG_EX_INT0] ),
              read_8250( dev->port[REG_EX_INT1] ), read_8250( dev->port[REG_EX_INT2] ),
              read_8250( dev->port[REG_EX_INT3] ) );

    devc_dbg( "[CH%d] 8xMode 0x%X 4xMode 0x%X", dev->chnum, read_8250( dev->port[REG_EX_8XMODE] ),
              read_8250( dev->port[REG_EX_4XMODE] ) );

    devc_dbg( "[CH%d] SLEEP 0x%X", dev->chnum, read_8250( dev->port[REG_EX_SLEEP] ) );

    devc_dbg( "[CH%d] DREV 0x%X DVID 0x%X", dev->chnum, read_8250( dev->port[REG_EX_DREV] ),
              read_8250( dev->port[REG_EX_DVID] ) );

    /* write_8250( dev->port[REG_EX_SLEEP], 0 ); */
    /* write_8250( dev->port[REG_EX_4XMODE], 0 ); */
    /* write_8250( dev->port[REG_EX_8XMODE], 0 ); */

    /* set_port( dev->port[REG_EFR], 1 << 4, 1 << 4 ); */
    /* set_port( dev->port[REG_IE], 1 << 6, 1 << 7 ); */
    /* set_port( dev->port[REG_MC], MCR_OUT1, MCR_OUT1 ); */

    devc_info( "[CH%d] Set device mode %s", dev->chnum,
               (exar_dev.mode[dev->chnum - exar_dev.ch_offset] == EXAR_MODE_RS232) ? "RS232" : "RS485" );
    
    fctr = FIELD_SET( 3, REG_FCTR_RTS_DTR_HYST ) | FIELD_SET( 3, REG_FCTR_TRG_TABLE );
    
    if ( exar_dev.mode[dev->chnum - exar_dev.ch_offset] == EXAR_MODE_RS485 )
        fctr |= 1 << REG_FCTR_AUTO_RS485_OFFSET;

    write_8250( dev->port[REG_FCTR], fctr );

    if ( exar_dev.mode[dev->chnum - exar_dev.ch_offset] == EXAR_MODE_RS485 && exar_dev.rs485_invert_polarity )
    {
        devc_info( "[CH%d] Invert RS485 Polarity", dev->chnum );
        set_port( dev->port[REG_LC], LCR_DLAB, LCR_DLAB );
        set_port( dev->port[REG_DLD], REG_DLD_INVERT_RS485_POLARITY, REG_DLD_INVERT_RS485_POLARITY );
        set_port( dev->port[REG_EFR], REG_EFR_ENABLE_SHARED_BITS, REG_EFR_ENABLE_SHARED_BITS );
        set_port( dev->port[REG_LC], LCR_DLAB, 0 );
    }
}


int enable_device( DEV_8250 *dev )
{
    unsigned char   reg = 0;

    /* Only setup IRQ handler for non-pcmcia devices. Pcmcia devices will have this done later when card is inserted. */
    if(dev->port != 0 && dev->intr != _NTO_INTR_SPARE)
    {
        /* If CREAD is set (default) then we must clear it before we call ser_stty(), else ser_stty() will
         * enable the receive interrupt before we attach our ISR. ser_attach_intr() will enable all the
         * interrupts after the ISR is attached. */
        if ( dev->tty.c_cflag & CREAD )
        {
            dev->tty.c_cflag &= ~(CREAD);
            ser_stty( dev );
            dev->tty.c_cflag |= CREAD;
        } else
            ser_stty( dev );

        if ( ser_attach_intr( dev ) == -1 )
        {
            free_device( dev );
            return (-1);
        }
    }
    /* rx and tx fifo sizes */
    dev->rx_fifo = EXAR_FIFO_RX_SIZE;
    dev->tx_fifo = EXAR_FIFO_TX_SIZE;

    devc_info( "[CH%d] RX FIFO: %d, TX FIFO: %d", dev->chnum, dev->rx_fifo, dev->tx_fifo );

    /* Set RX fifo trigger level */
    switch ( dev->tty.fifo & 0xf )
    {
        case 1:
        default:
            reg = FCR_RX_TRIG_1;
            break;
        case 4:
            reg = FCR_RX_TRIG_4;
            break;
        case 8:
            reg = FCR_RX_TRIG_8;
            break;
        case 14:
            reg = FCR_RX_TRIG_14;
            break;
    }

    write_8250( dev->port[REG_RXTRG], dev->rx_fifo );
    write_8250( dev->port[REG_TXTRG], dev->tx_fifo );

    reg |= FCR_FIFO_ENABLE | FCR_RX_FIFO_RESET | FCR_TX_FIFO_RESET;
    write_8250( dev->port[REG_FC], reg );

    /* Turn on DTR and RTS */
    set_port( dev->port[REG_MC], MCR_DTR | MCR_RTS, MCR_DTR | MCR_RTS );

    /* Get current MSR stat */
    reg = read_8250( dev->port[REG_MS] );

    if ( reg & MSR_DDCD )
        tti( &dev->tty, (reg & MSR_DCD) ? TTI_CARRIER : TTI_HANGUP );

    if ( (reg & MSR_DCTS) && (dev->tty.c_cflag & OHFLOW) )
        tti( &dev->tty, (reg & MSR_CTS) ? TTI_OHW_CONT : TTI_OHW_STOP );

    enable_exar_ch( dev );

    /* Attach the resource manager */
    ttc( TTC_INIT_ATTACH, &dev->tty, 0 );

    return (EOK);
}
