/*
 * (c) 2022, SWD Embedded Systems Limited, http://www.kpda.ru
 */


#include "externs.h"


/* Process data in a line status register */
unsigned process_lsr( DEV_8250 *dev, unsigned char lsr )
{
    unsigned    key = 0;    /*, eventflag = 0; */

    /* Return immediately if no errors. */
    if ( (lsr & (LSR_BI | LSR_OE | LSR_FE | LSR_PE)) == 0 )
        return (0);

    /* Save the error as out-of-band data which can be retrieved via devctl(). */
    dev->tty.oband_data |= (lsr >> 1) & 0x0f;
    atomic_set( &dev->tty.flags, OBAND_DATA );

    if ( lsr & LSR_BI )
        key |= TTI_BREAK;
    else {
        if ( lsr & LSR_OE )
            key |= TTI_OVERRUN;
        else if ( lsr & LSR_FE )
            key |= TTI_FRAME;
        else if ( lsr & LSR_PE )
            key |= TTI_PARITY;
    }

    return (key);
}


/* Serial interrupt handler */
const struct sigevent * ser_intr( void *area, int id )
{
    struct dev_list *list  = area;
    int             status = 0;
    int             something_happened;
    unsigned char   msr,
                    lsr;
    DEV_8250        *dev;
    struct sigevent *event = NULL;
    unsigned        c;
    int             channel;

    do {
        something_happened = 0;
        for ( dev = list->device; dev != NULL; dev = dev->next )
        {
            unsigned    iir,
                        rxcnt = 0;
            uintptr_t   *port = dev->port;

            status = 0;

            iir = (read_8250( port[REG_II] ) & 0x07);

            switch ( iir )
            {
                case II_RX:     /* Receive data */
                case II_LS:
                    /* Some UARTs will generate a LS interrupt when there is an error anywhere in the RX
                     * FIFO, and will clear this interrupt only when there are no more errors remaining
                     * in the FIFO. The error bits in REG_LS (BI/PR/FE/OE) always represent the error
                     * status for the received character at the top of the Rx FIFO. Reading the Rx FIFO
                     * updates these bits to the appropriate status of the new character. This means that
                     * it is possible to get an LS interrupt with none of the error status bits set, in
                     * order to clear the LS interrupt we must read out all of the characters in the FIFO
                     * until we find and handle the erronous character. */
                    if ( (lsr = read_8250( port[REG_LS] )) & LSR_RXRDY )
                    {
                        c        = process_lsr( dev, lsr );
                        rxcnt    = read_8250( port[REG_RXCNT] );
                        channel  = dev->chnum - exar_dev.ch_offset - 1;
                        status  |= tti2( &dev->tty, dev->parent_dev->uart_mem + UART_READ_FIFO_BASE( channel ), rxcnt, c );
                    };
                    break;

                case II_TX:     /* Transmit buffer empty */
                    dev->tty.un.s.tx_tmr = 0;
                    /* Send evnet to io-char, tto() will be processed at thread time */
                    atomic_set( &dev->tty.flags, EVENT_TTO );
                    status |= 1;
                    break;

                case II_MS:     /* Modem change */
                    msr = read_8250( port[REG_MS] );
                    if ( msr & MSR_DDCD )
                        status |= tti( &dev->tty, (msr & MSR_DCD) ? TTI_CARRIER : TTI_HANGUP );

                    if ( (msr & MSR_DCTS) && (dev->tty.c_cflag & OHFLOW) )
                        status |= tti( &dev->tty, (msr & MSR_CTS) ? TTI_OHW_CONT : TTI_OHW_STOP );

                    /* OBAND notification of Modem status change */
                    dev->tty.oband_data |= _OBAND_SER_MS;
                    atomic_set( &dev->tty.flags, OBAND_DATA );
                    atomic_set( &dev->tty.flags, EVENT_NOTIFY_OBAND );
                    status |= 1;
                    break;

                default:
                    continue;
            }

            something_happened = 1;
            if ( status )
                if ((dev->tty.flags & EVENT_QUEUED) == 0 )
                {
                    event = &ttyctrl.event;
                    dev_lock( &ttyctrl );
                    ttyctrl.event_queue[ttyctrl.num_events++] = &dev->tty;
                    atomic_set( &dev->tty.flags, EVENT_QUEUED );
                    dev_unlock( &ttyctrl );
                }
        }
    } while ( something_happened );

    return (event);
}
