/*
 * (c) 2022, SWD Embedded Systems Limited, http://www.kpda.ru
 */


/*
#ifdef __USAGE
%C - Serial driver for PCIe cards based on Exar XR17V354/358 and compatible controllers

%C [options]
Options:
 -b number       Define initial baud rate (default 115200)
 -c clk[/div]    Set the input clock rate and divisor
 -C number       Size of canonical input buffer (default 1024)
 -e              Set options to "edit" mode
 -E              Set options to "raw" mode (default)
 -I number       Size of raw input buffer (default 2048)
 -f              Enable hardware flow control (default)
 -F              Disable hardware flow control
 -O number       Size of output buffer (default 2048)
 -s              Enable software flow control
 -S              Disable software flow control (default)
 -t number       Set receive FIFO trigger level (default 14)
 -u unit         Set serial unit number (default 1)
 -o opt[,opt]    String options:
                 pmm_parent=path (pmm_parent power pathname, default NULL)
                 pm_noflw (Disable flow control on power down, default enabled)
                 highwater=value (RX watermark for input flow control (bytes))
                 disable=rx (Disable receiver on startup)
 -v[vvv]         Increase verbosity level: errors, warnings, info, debug
 -V              PCI Vendor ID (0x13a8 by default)
 -D              PCI Device ID (0x358 by default)
 -i              PCI index
 -R              Interrupt number
 -P              Number of channels (default 8)
 -m RS485|RS232  Set selected mode for all ports (RS232 by default)
 -M port=1[,port=2] Set RS485/422 mode for specific ports
 -p              Invert RS485 Polarity
 -r              Disable DTR control
 -a              Enable ASPM (default: disable)

Examples:
Default RS232 mode
%C

RS485 mode with disabled hardware flow control
%C -mRS485 -F

RS485 mode with inverted polarity for Advantech PCI-1612 0x13fe:0x1b PCI card
%C -mRS485 -v -p -V 0x13fe -D 0x1b -i 0 
#endif
*/


#include "externs.h"


unsigned options( int argc, char *argv[] )
{
    int             opt,
                    i,
                    enabled_cnt = 0,
                    numports    = 0,
                    port        = -1;
    char            *cp;
    char            *options,
                    *value;
    char            *port_options[] = { "port", NULL };
    unsigned        unit,
                    dtr_opt  = 0;
    unsigned        fifo_rx;
    uint32_t        chsize   = EXAR_UART_CHANNEL_SIZE;
    uint32_t        choffset = EXAR_UART_CHANNEL_OFFSET;
    DEV_8250        *dev_list[MAX_DEVICES];
    int32_t         pci_id,
                    pci_idx = 0;
    void            *pci_handle;
    static TTYINIT  devinit  = { 0, 0, 0, 115200,                   /* port port_shift intr baud */
                                 4096, 4096, 1024,                  /* isize osize csize */
                                 0, 0, 0, 0, 0xee, 125000000, 16,   /* c_cflag c_iflag c_lflag c_oflag fifo clk div */
                                 "/dev/ser",
                                 NULL, 0, 0, 0 };
    struct pci_dev_info inf;

    if ( (pci_id = pci_attach( 0 )) == -1 )
    {
        devc_err( "io-char: Unable to attach to PCI server (%d)", errno );
        return (-1);
    }

    /* Exar PCI device */
    inf.VendorId       = EXAR_VID;
    inf.DeviceId       = EXAR_XR17D158_DID;
    exar_dev.uart_cnt  = -1;
    exar_dev.ch_offset = -1;
    for ( i = 1; i <= UART_MAX_CH_COUNT; i++ )
        exar_dev.mode[i] = EXAR_MODE_RS232;
    exar_dev.disable_aspm          = 1;
    exar_dev.rs485_invert_polarity = 0;

    /* Initialize the devinit to raw mode */
    ttc( TTC_INIT_RAW, &devinit, 0 );

    sys_ttyinit( &devinit );
    unit = 1;

    while ( optind < argc )
    {
        /* Process options */
        while ( (opt = getopt( argc, argv, IO_CHAR_SERIAL_OPTIONS "c:t:T:u:i:V:D:R:P:m:M:arp" )) != -1 )
        {
            switch ( ttc( TTC_SET_OPTION, &devinit, opt ) )
            {
                case 'c':
                    devinit.clk = strtoul( optarg, &optarg, 0 );
                    if ( (cp = strchr( optarg, '/' )) )
                        devinit.div = strtoul( cp + 1, NULL, 0 );
                    break;

                case 't':
                    fifo_rx = strtoul( optarg, NULL, 0 );
                    if ( !((fifo_rx == 1) || (fifo_rx == 4) || (fifo_rx == 8) || (fifo_rx == 14)) )
                    {
                        devc_err( "Illegal rx fifo trigger. \n" );
                        devc_err( "Trigger number must be 1, 4, 8 or 14. \n" );
                        devc_err( "Rx trigger will not be enabled.\n\n" );
                        fifo_rx = 0;
                    } else {
                        devinit.fifo &= 0xf0;
                        devinit.fifo |= fifo_rx;
                    }
                    break;

                case 'u':
                    unit = strtoul( optarg, NULL, 0 );
                    break;

                /* PCI index */
                case 'i':
                    pci_idx = strtoul( optarg, NULL, 0 );
                    break;

                /* PCI Vendor */
                case 'V':
                    inf.VendorId = strtoul( optarg, NULL, 0 );
                    break;

                /* PCI device */
                case 'D':
                    inf.DeviceId = strtoul( optarg, NULL, 0 );
                    break;

                /* Number of ports */
                case 'P':
                    exar_dev.uart_cnt = strtoul( optarg, NULL, 0 );
                    break;

                /* IRQ */
                case 'R':
                    devinit.intr = strtoul( optarg, NULL, 0 );
                    break;

                /* RS232/RS485 mode */
                case 'm':
                    if ( strcasecmp( "RS485", optarg ) == 0 )
                        for ( i = 1; i <= UART_MAX_CH_COUNT; i++ )
                            exar_dev.mode[i] = EXAR_MODE_RS485;
                    else if ( strcasecmp( "RS232", optarg ) == 0 )
                        for ( i = 1; i <= UART_MAX_CH_COUNT; i++ )
                            exar_dev.mode[i] = EXAR_MODE_RS232;
                    else
                        devc_err( "Error in device mode string parsing('%s'). Use default", optarg );
                    break;

                /* RS485 per port options */
                case 'M':
                    options = optarg;
                    while ( *options != '\0' )
                    {
                        switch ( getsubopt( &options, port_options, &value ) )
                        {
                            case 0:
                                if ( value != NULL )
                                {
                                    port = atoi( value );   /* 1-8 */
                                    if ( port > 0 && port <= UART_MAX_CH_COUNT )
                                    {
                                        exar_dev.mode[port] = EXAR_MODE_RS485;
                                        devc_info( "Port[%d] mode: RS485/422", port );
                                    } else
                                        devc_err( "Incorrect port mode option. Valid port numbers: 1-%d", UART_MAX_CH_COUNT );
                                }
                                break;
                            default :
                                break;
                        }
                    }
                    break;

                /* Enable Active-state power managment */
                case 'a':
                    exar_dev.disable_aspm = 0;
                    break;

                /* Disable DTR control. Workaround for Qt QtSerialPort class */
                case 'r':
                    dtr_opt = 1;
                    break;

                /* RS485 inverted polarity */
                case 'p':
                    exar_dev.rs485_invert_polarity = 1;
                    break;
            }
        }

        /* Process ports and interrupts */
        while ( optind < argc  &&  *(optarg = argv[optind]) != '-' )
        {
            devinit.port = strtoull( optarg, &optarg, 16 );
            choffset     = devinit.port;

            if ( *optarg == ':' )
                chsize = strtoull( optarg + 1, &optarg, 0 );
            if ( *optarg == '^' )
                devinit.port_shift = strtoul( optarg + 1, &optarg, 0 );
            if ( *optarg == ',' )
                devinit.intr = strtoul( optarg + 1, NULL, 0 );
            ++optind;
        }
    }
    driver_verbosity_level = devinit.verbose;

    if ( pci_idx != -1 )
    {
        devc_dbg( "Attach to PCI device vid:0x%X did:0x%X", inf.VendorId, inf.DeviceId );

        if ( (pci_handle = pci_attach_device( NULL, PCI_SEARCH_VENDEV | PCI_MASTER_ENABLE | PCI_INIT_ALL,
                                              pci_idx, &inf )) == NULL )
        {
            pci_detach( pci_id );
            devc_err( "io-char: Can't attach to PCI device 0x%X,0x%X.", inf.VendorId, inf.DeviceId );
            return (-1);
        }
    }

    if ( exar_dev_init( &inf, &exar_dev ) < 0 )
    {
        pci_detach( pci_id );
        devc_err( "Failed to init Exar device." );
        fprintf( stderr, "Failed to init Exar device\n" );
        return (-1);
    }

    /* Process ports and interrupts. */
    for ( i = 0; i < exar_dev.uart_cnt; i++ )
    {
        /* Offset for Exar chip */
        devinit.port = PCI_MEM_ADDR( inf.CpuBaseAddress[0] ) + (choffset + (chsize * i));
        devc_dbg( "Run driver on channel %d. Mem: %llX (base offset : 0x%X , channel offset : 0x%X)", i, devinit.port, choffset, chsize );

        if ( devinit.intr == 0 )
            devinit.intr = inf.Irq;

        if ( (dev_list[numports] = create_device( &devinit, unit++, &exar_dev )) == NULL )
            devc_err( "io-char: Initialization of /dev/ser%d (port 0x%llx) failed", unit - 1, devinit.port );
        else
            ++numports;
    }

    if ( numports == 0 )
    {
        devc_err( "Failed to init Exar device." );
        fprintf( stderr, "Failed to init Exar device\n" );
        return (-1);
    }

    /* Enable all ports (Attach ISR and pathname entry) */
    for ( i = 0; i < numports; i++ )
    {
        if ( dev_list[i]->tty.verbose )
        {
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Port .......................... %s (0x%x)", dev_list[i]->tty.name, dev_list[i]->port[0] );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "IRQ ........................... 0x%x", dev_list[i]->intr );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Input Clock ................... %d", dev_list[i]->clk );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Clock Divisor ................. %d", dev_list[i]->div );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Input buffer size ............. %d", dev_list[i]->tty.ibuf.size );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Input flow control highwater .. %d", dev_list[i]->tty.highwater );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Output buffer size ............ %d", dev_list[i]->tty.obuf.size );
            slogf( _SLOG_SETCODE( _SLOGC_CHAR, 0 ), _SLOG_INFO, "Canonical buffer size ......... %d\n", dev_list[i]->tty.cbuf.size );
        }
        if ( enable_device( dev_list[i] ) != -1 )
        {
            enabled_cnt++;
            dev_list[i]->dtr_disable = dtr_opt;
        }
    }

    return (enabled_cnt);
}
